from numpy import linalg
from pytplot import split_vec, join_vec, get_data, store_data, options
from typing import Union, List
import logging

def _tvectot(tvar: str, new_name: str, join_component: bool):
    data = get_data(tvar)
    md = get_data(tvar,metadata=True)
    new_data = linalg.norm(data.y, axis=1)
    store_data(new_name, data={'x':data.times,'y':new_data},attr_dict=md)
    
    if join_component:
        # Caution: this can clobber existing variable names!  Maybe better to manipulate the
        # data directly instead of calling split_vec/join_vec
        join_vec(split_vec(tvar)+[new_name], newname=new_name)
        options(new_name, 'legend_names', ['x', 'y', 'z', 'Magnitude'])
    else:
        options(new_name, 'legend_names', 'Magnitude')
    return new_name

def tvectot(tvars: Union[str, List[str]], newname=None, newnames: Union[str, List[str]] = None, suffix="_mag", join_component=False) -> Union[str , List[str]]:
    """
    Computes the magnitude of a vector time series.

    Parameters
    ----------
    - tvars : Names of the tplot variables.
    - newnames: (Deprecated) Names for the resultant magnitude tplot variables. If not provided, it appends the suffix to `tvars`.
    - newname: Names for the resultant magnitude tplot variables. If not provided, it appends the suffix to `tvars`.
    - suffix: The suffix to append to tensor_names to form new_names if new_names is not provided.
    - join_component: If True, the magnitude tplot variable is joined with the component tplot variables.

    Returns
    -------
    Names of the magnitude tplot variables.
    """

    # newnames is deprecated in favor of newname
    if newnames is not None:
        logging.info("tvectot: The newnames parameter is deprecated. Please use newname instead.")
        newname = newnames

    tvars_type = type(tvars)
    if tvars_type == str:
        tvars = [tvars]
    if join_component:
        suffix = "_tot"

    if newname is None:
        newname = [tvar + suffix for tvar in tvars]

    # If newname is a scalar string, zip will iterate over characters
    if type(newname) is not list:
        newname = [newname]

    for tvar, newname_single in zip(tvars, newname):
        _tvectot(tvar, newname_single, join_component)
    
    if tvars_type == str:
        return newname[0]
    else:
        return newname



# Import pyspedas tools into pyspedas namespace
# These are imported as a convenience for pyspedas users.  For internal pyspedas development, it is
# probably best to keep using the fully qualified module names.

from .analysis.avg_data import avg_data
from .analysis.deriv_data import deriv_data
from .analysis.tvectot import tvectot
from .analysis.tinterpol import tinterpol
from .analysis.yclip import yclip
from .analysis.twavpol import twavpol
from .analysis.wavelet import wavelet
from .analysis.time_domain_filter import time_domain_filter
from .analysis.find_magnetic_nulls import find_magnetic_nulls_fote, classify_null_type
from .analysis.lingradest import lingradest
from .cdagui_tools.cdagui import cdagui
from .cdagui_tools.cdaweb import CDAWeb
from .cotrans_tools.cotrans import cotrans
from .cotrans_tools.cotrans_get_coord import cotrans_get_coord
from .cotrans_tools.cotrans_set_coord import cotrans_set_coord
from .cotrans_tools.tvector_rotate import tvector_rotate
from .cotrans_tools.cart2spc import cart2spc
from .cotrans_tools.spc2cart import spc2cart
from .cotrans_tools.sm2mlt import sm2mlt
from .cotrans_tools.fac_matrix_make import fac_matrix_make
from .cotrans_tools.gsm2lmn import gsm2lmn
from .cotrans_tools.minvar import minvar
from .cotrans_tools.minvar_matrix_make import minvar_matrix_make
from .cotrans_tools.quaternions import qtom, qconj, qdotp, qmult, qnorm, qslerp, qcompose, qvalidate, qdecompose, mtoq, qnormalize
from .cotrans_tools.tvector_rotate import tvector_rotate
from .cotrans_tools.xyz_to_polar import xyz_to_polar
# Importing geopack causes IGRF coefficients to be loaded by the external geopack package, which may not be desired.
#from .geopack.get_tsy_params import get_tsy_params
#from .geopack.get_w_params import get_w
#from .geopack.kp2iopt import kp2iopt
#from .geopack.t01 import t01, tt01
#from .geopack.t89 import t89, tt89
#from .geopack.t96 import t96, tt96
#from .geopack.ts04 import tts04
from .hapi_tools.hapi import hapi
from .projects.noaa.noaa_load_kp import noaa_load_kp
from .particles.moments import moments_3d, spd_pgs_moments, spd_pgs_moments_tplot
from .particles.spd_part_products import spd_pgs_do_fac, spd_pgs_regrid
from .particles.spd_slice2d import slice1d_plot, slice2d, slice2d_plot
from .utilities.spice.time_ephemeris import time_ephemeris
from .utilities.dailynames import dailynames
from .utilities.datasets import find_datasets
# Note: "download" and "download_file" might be problematic names to import, due to risk of conflict with other packages
from .utilities.download import download, download_file, check_downloaded_file
from .utilities.download_ftp import download_ftp
from .utilities.find_ip_address import find_ip_address
from .utilities.interpol import interpol
from .utilities.leap_seconds import load_leap_table
from .utilities.libs import libs
from .utilities.mpause_2 import mpause_2
from .utilities.mpause_t96 import mpause_t96
from .utilities.tcopy import tcopy
from .version import version


# Import pytplot tools into pyspedas namespace
# Note to developers: Do not use these imports for pyspedas internals, or it may cause
# circular dependencies.  Import directly from pytplot instead.

from pytplot import *

# omni must precede mms to avoid problems with circular imports
from .projects import omni

# Import routine names with mission prefixes into pyspedas namespace
from .projects.mms import mms_load_mec, mms_load_fgm, mms_load_scm, mms_load_edi, \
    mms_load_edp, mms_load_eis, mms_load_feeps, \
    mms_load_hpca, mms_load_fpi, mms_load_aspoc, \
    mms_load_dsp, mms_load_fsm, mms_load_state, \
    mms_qcotrans, mms_cotrans_lmn, mms_cotrans_qrotate, mms_cotrans_qtransformer
from .projects.mms.feeps.mms_feeps_pad import mms_feeps_pad
from .projects.mms.feeps.mms_feeps_gpd import mms_feeps_gpd
from .projects.mms.eis.mms_eis_pad import mms_eis_pad
from .projects.mms.hpca.mms_hpca_calc_anodes import mms_hpca_calc_anodes
from .projects.mms.hpca.mms_hpca_spin_sum import mms_hpca_spin_sum
from .projects.mms.plots.mms_overview_plot import mms_overview_plot
from .projects.mms.particles.mms_part_getspec import mms_part_getspec
from .projects.mms.particles.mms_part_slice2d import mms_part_slice2d


# The code below is needed for backward compatibility, so users can continue to do things
# like "from pyspedas.mms import mec" even after mms has been moved to the projects directory.

import sys
from importlib import import_module

# List of submodules we want to make available under the pyspedas namespace
submodules = ['ace', 'akebono', 'barrel', 'cluster', 'cnofs', 'csswe', 'de2', 'dscovr',
             'elfin', 'equator_s', 'erg', 'fast', 'geotail', 'goes', 'image', 'kompsat',
              'kyoto', 'lanl', 'maven', 'mica', 'mms', 'noaa', 'omni', 'poes', 'polar', 'psp',
              'rbsp', 'secs', 'soho', 'solo', 'st5', 'stereo', 'swarm', 'themis', 'themis.state', 'twins',
              'ulysses'
              ]

for submodule in submodules:
    # Import the module from the new path
    full_module_path = f"pyspedas.projects.{submodule}"
    imported_module = import_module(full_module_path)

    # Add it to sys.modules under the old path
    sys.modules[f"pyspedas.{submodule}"] = imported_module

# This set of imports is still needed for backward compatibility, when using fully-qualified
# routine names in function calls, like "pyspedas.mms.mec()" rather than "pyspedas.projects.mms.mec()"

# Make mission-specific namespaces available under pyspedas
from .projects import ace
from .projects import akebono
from .projects import barrel
from .projects import cluster
from .projects import cnofs
from .projects import csswe
from .projects import de2
from .projects import dscovr
from .projects import elfin
from .projects import equator_s
from .projects import erg
from .projects import fast
from .projects import geotail
from .projects import goes
from .projects import image
from .projects import kompsat
# for backward compatibility
from .projects.kompsat.load import load as kompsat_load
from .projects import kyoto
from .projects import lanl
from .projects import maven
# for backward compatibility
from .projects.maven import maven_load
from .projects import mica
from .projects import mms
from .projects import noaa
from .projects import poes
from .projects import polar
from .projects import psp
from .projects import rbsp
from .projects import secs
from .projects import soho
from .projects import solo
from .projects import st5
from .projects import stereo
from .projects import swarm
from .projects import themis
from .projects import twins
from .projects import ulysses
from . import vires
from .projects import wind

# set up logging/console output
import logging
from os import environ

logging_level = environ.get('PYSPEDAS_LOGGING_LEVEL')
logging_format = environ.get('PYSPEDAS_LOGGING_FORMAT')
logging_date_fmt = environ.get('PYSPEDAS_LOGGING_DATE_FORMAT')

if logging_format is None:
    logging_format = '%(asctime)s: %(message)s'

if logging_date_fmt is None:
    logging_date_fmt = '%d-%b-%y %H:%M:%S'

if logging_level is None:
    logging_level = logging.INFO
else:
    logging_level = logging_level.lower()
    if logging_level == 'debug':
        logging_level = logging.DEBUG
    elif logging_level == 'info':
        logging_level = logging.INFO
    elif logging_level == 'warn' or logging_level == 'warning':
        logging_level = logging.WARNING
    elif logging_level == 'error':
        logging_level = logging.ERROR
    elif logging_level == 'critical':
        logging_level = logging.CRITICAL

logging.captureWarnings(True)

# basicConfig here doesn't work if it has previously been called
logging.basicConfig(format=logging_format, datefmt=logging_date_fmt, level=logging_level)

# manually set the logger options from the defaults/environment variables
logger = logging.getLogger()
logger_handler = logger.handlers[0]  # should exist since basicConfig has been called
logger_fmt = logging.Formatter(logging_format, logging_date_fmt)
logger_handler.setFormatter(logger_fmt)
logger.setLevel(logging_level)



from pyspedas.projects.mms.mms_load_data import mms_load_data



def mms_load_fsm(trange=['2015-10-16/05:59', '2015-10-16/06;01'], probe='1', data_rate='brst',
    level='l3', datatype='8khz', get_support_data=False, time_clip=False, no_update=False, 
    available=False, varformat=None, varnames=[], notplot=False, suffix='', latest_version=False, 
    major_version=False, min_version=None, cdf_version=None, spdf=False, always_prompt=False):
    """
    This function loads MMS FSM (FGM + SCM) data into tplot variables
    
    Parameters
    ----------
        trange : list of str
            time range of interest [start time, end time] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2015-10-16/05:59', '2015-10-16/06:01']

        probe : str or list of str
            list of probes, valid values for MMS probes are ['1','2','3','4'].
            Default: '1'

        data_rate : str or list of str
            Valid values: 'brst'
            Default: 'brst'

        level : str
            indicates level of data processing.
            Default: 'l3'

        datatype : str or list of str
            Valid datatype for FSM is: 8khz
            Default: '8khz'

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".
            Default: False

        time_clip: bool
            Data will be clipped to the exact trange specified by the trange keyword.
            Default: False

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.
            Default: None (all variables are loaded)

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)
            Default: [] (all variables are loaded)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.
            Default: None

        notplot: bool
            If True, then data are returned in a hash table instead of 
            being stored in tplot variables (useful for debugging, and
            access to multidimensional data products)
            Default: False

        available: bool
            If True, simply return the available data files (without downloading)
            for the requested parameters
            Default: False

        no_update: bool
            Set this flag to preserve the original data. if not set and newer 
            data is found the existing data will be overwritten
            Default: False

        cdf_version: str
            Specify a specific CDF version # to load (e.g., cdf_version='4.3.0')
            Default: None

        min_version: str
            Specify a minimum CDF version # to load
            Default: None

        latest_version: bool
            Only grab the latest CDF version in the requested time interval
            Default: False

        major_version: bool
            Only open the latest major CDF version (e.g., X in vX.Y.Z) in the requested time interval
            Default: False

        always_prompt: bool
            Set this keyword to always prompt for the user's username and password;
            useful if you accidentally save an incorrect password, or if your SDC password has changed
            Default: False
            
        spdf: bool
            If True, download the data from the SPDF instead of the SDC
            Default: False

    Returns
    -------
        list of str
            List of tplot variables created.

    Example
    -------

    >>> import pyspedas
    >>> from pytplot import tplot
    >>> fsm_data = pyspedas.projects.mms.mms_load_fsm(trange=['2015-10-16/05:59', '2015-10-16/06:01'])
    >>> tplot(['mms1_fsm_b_mag_brst_l3','mms1_fsm_b_gse_brst_l3','mms1_fsm_r_gse_brst_l3'])

    """
    tvars = mms_load_data(trange=trange, notplot=notplot, varformat=varformat, probe=probe, data_rate=data_rate, 
        level=level, instrument='fsm', datatype=datatype, get_support_data=get_support_data, time_clip=time_clip, 
        no_update=no_update, available=available, suffix=suffix, latest_version=latest_version, varnames=varnames,
        major_version=major_version, min_version=min_version, cdf_version=cdf_version, spdf=spdf, always_prompt=always_prompt)
    return tvars



"""Test functions in the utilites folder."""
import unittest
import pyspedas
from copy import deepcopy

from pytplot import del_data, tplot, tplot_options, tplot_names
from pytplot import store_data, get_data
from pytplot import wildcard_expand, tname_byindex, tplot_wildcard_expand, tindex_byname

global_display=False
tp_data={'x':[1,2,3], 'y':[4,5,6]}

class UtilTestCases(unittest.TestCase):
    """Test functions related to tplot variable indexing and wildcards."""

    def test_wildcard_expand(self):
        # Test wildcard expansion against defined list
        master_list = ['Foo', 'foo', 'tha_pos_gsm', 'thb_pos_gsm', 'mms1_mec_r_sm']
        # Literal match
        result = wildcard_expand(master_list,"Foo")
        self.assertEqual(result,['Foo'])
        # Results should be ordered following patterns, not master list
        result = wildcard_expand(master_list,['foo', 'Foo'])
        self.assertEqual(result, ['foo', 'Foo'])
        # Wildcard expansion
        result = wildcard_expand(master_list,"th?_*_*")
        self.assertEqual(result,['tha_pos_gsm', 'thb_pos_gsm'])
        # Case sensitive (by default), no matches found
        result = wildcard_expand(master_list,"*SM")
        self.assertEqual(result, [])
        # Case insensitive
        result = wildcard_expand(master_list,"*SM", case_sensitive=False)
        self.assertEqual(result, ['tha_pos_gsm', 'thb_pos_gsm', 'mms1_mec_r_sm'])
        # Null pattern
        result = wildcard_expand(master_list, None)
        self.assertEqual(result, [])
        # Master list and pattern both null
        result = wildcard_expand(None,None)
        self.assertEqual(result, [])
        # Master list not a list
        result = wildcard_expand('foo', 'foo')
        self.assertEqual(result, [])

    def test_tplot_wildcard_expand(self):
        # Test wildcard expansion against current tplot variables
        del_data('*')
        master_list = ['Foo', 'foo', 'tha_pos_gsm', 'thb_pos_gsm', 'mms1_mec_r_sm']
        for var in master_list:
            store_data(var,deepcopy(tp_data))
        # Add a composite variable
        store_data('composite',data=['foo', 'tha_pos_gsm', 'thb_pos_gsm'])
        tplot_names()
        tn = tplot_names(quiet=True)
        print(tn)
        # Literal match
        result = tplot_wildcard_expand("Foo")
        self.assertEqual(result,['Foo'])
        # Results should be ordered following patterns, not master list
        result = tplot_wildcard_expand(['foo', 'Foo'])
        self.assertEqual(result, ['foo', 'Foo'])
        # Wildcard expansion
        result = tplot_wildcard_expand("th?_*_*")
        self.assertEqual(result,['tha_pos_gsm', 'thb_pos_gsm'])
        # Composite variable
        result = tplot_wildcard_expand('composite')
        self.assertEqual(result, ['composite'])
        # Case sensitive (by default), no matches found
        result = tplot_wildcard_expand("*SM")
        self.assertEqual(result, [])
        # Case insensitive
        result = tplot_wildcard_expand("*SM", case_sensitive=False)
        self.assertEqual(result, ['tha_pos_gsm', 'thb_pos_gsm', 'mms1_mec_r_sm'])
        # Mixed strings and integers with repeated values
        result = tplot_wildcard_expand(['Foo',1,2,2,'tha*'], case_sensitive=False)
        self.assertEqual(result, ['Foo','foo','tha_pos_gsm'])
        # Null pattern
        result = tplot_wildcard_expand(None)
        self.assertEqual(result, [])

    def test_tindex_byname(self):
        # Test lookups of tplot variable indices by name
        del_data('*')
        store_data('foo',data=deepcopy(tp_data))
        store_data('bar',data=deepcopy(tp_data))
        result=tindex_byname('foo')
        self.assertEqual(result,0)
        result=tindex_byname('bar')
        self.assertEqual(result,1)
        result=tindex_byname('doesnt_exist')
        self.assertEqual(result,None)

    def test_tname_byindex(self):
        # Test lookups of tplot variable names by indices
        del_data('*')
        # Out of bounds (no tplot variables yet)
        result=tname_byindex(0)
        self.assertEqual(result,None)
        store_data('foo',data=deepcopy(tp_data))
        store_data('bar',data=deepcopy(tp_data))
        result=tname_byindex(0)
        self.assertEqual(result,'foo')
        result=tname_byindex(1)
        self.assertEqual(result,'bar')
        # Index too large
        result=tname_byindex(2)
        self.assertEqual(result,None)
        # Negative index
        result=tname_byindex(-1)
        self.assertEqual(result,None)


    def test_plotting_wildcards_indices(self):
        # Test tplot using wildcards
        del_data('*')
        pyspedas.projects.themis.state(probe=['a', 'b', 'c', 'd', 'e'])
        # Wildcard pattern
        tplot_options('title',"Positions of THEMIS A B C D E")
        tplot('th?_pos', save_png='tplot_wildcard.png', display=global_display)
        # List of integer indices
        indices = [tindex_byname('tha_pos'), tindex_byname('thb_pos'), tindex_byname('thc_pos')]
        tplot_options('title',"Positions of THEMIS A B C")
        tplot(indices,save_png='tplot_byindex.png', display=global_display)
        # List including a bad name
        tplot_options('title',"Positions of THEMIS A B C D E")
        tplot(['th?_pos', 'doesnt_exist'], save_png='tplot_list_onebad.png', display=global_display)


if __name__ == '__main__':
    unittest.main()




import os
import unittest

from pyspedas.utilities.download import download

class DownloadTestCases(unittest.TestCase):
    def test_remote_path(self):
        # only specifying remote_path saves the files to the current working directory
        files = download(remote_path='https://spdf.gsfc.nasa.gov/pub/data/psp/sweap/spc/l3/l3i/2019/psp_swp_spc_l3i_20190401_v01.cdf')
        self.assertTrue(len(files) == 1)
        self.assertTrue(files[0] == os.path.join(os.getcwd(), 'psp_swp_spc_l3i_20190401_v01.cdf'))

    def test_remote_file(self):
        # only specifying remote_file saves the files to the current working directory
        files = download(remote_file='https://spdf.gsfc.nasa.gov/pub/data/psp/sweap/spc/l3/l3i/2019/psp_swp_spc_l3i_20190401_v01.cdf')
        self.assertTrue(len(files) == 1)
        self.assertTrue(files[0] == os.path.join(os.getcwd(), 'psp_swp_spc_l3i_20190401_v01.cdf'))

    def test_wildcard(self):
        # Test a wildcard pattern with several matches
        files = download(remote_path='http://themis.ssl.berkeley.edu/data/themis/tha/l1/state/2008/tha_l1_state_20080323_v??.cdf')
        self.assertTrue(len(files) == 4)  # v00, v01, v02, v03 should be available on this date
        self.assertTrue(files[3] == os.path.join(os.getcwd(), 'tha_l1_state_20080323_v03.cdf'))

    def test_missing_index(self):
        # Test a wildcard pattern on a nonexistent directory
        # This should warn "Remote index not found"
        files = download(remote_path='http://themis.ssl.berkeley.edu/data/themis/tha/l1/state/2006/tha_l1_state_20060323_v??.cdf')
        self.assertTrue(len(files) == 0)

    def test_last_version(self):
        # Test a wildcard pattern with several matches, and last_version=True, returning the final (lexicographic) value
        files = download(remote_path='http://themis.ssl.berkeley.edu/data/themis/tha/l1/state/2008/tha_l1_state_20080323_v??.cdf',last_version=True)
        self.assertTrue(len(files) == 1)  # v00, v01, v02, v03 should be available on this date, should only return v03
        self.assertTrue(files[0] == os.path.join(os.getcwd(), 'tha_l1_state_20080323_v03.cdf'))

    def test_last_version_nomatch(self):
        # Test a wildcard pattern that doesn't match anything.
        # This should warn about no matching file found in the index.
        files = download(remote_path='http://themis.ssl.berkeley.edu/data/themis/tha/l1/state/2008/tha_l1_state_20080332_v??.cdf',last_version=True)
        self.assertTrue(len(files) == 0)  # Nonexistent date, nothing should be found

    def test_remote_path_file(self):
        # specifying both remote_path and remote_file saves the files to the current working directory + the path specified in remote_file
        files = download(remote_path='https://spdf.gsfc.nasa.gov/pub/data/', remote_file='psp/sweap/spc/l3/l3i/2019/psp_swp_spc_l3i_20190401_v01.cdf')
        self.assertTrue(len(files) == 1)
        self.assertTrue(files[0] == os.path.join(os.getcwd(), 'psp/sweap/spc/l3/l3i/2019/psp_swp_spc_l3i_20190401_v01.cdf'))

    def test_local_file(self):
        # specifying local_file changes the local file name
        files = download(local_file='psp_data.cdf', remote_file='https://spdf.gsfc.nasa.gov/pub/data/psp/sweap/spc/l3/l3i/2019/psp_swp_spc_l3i_20190401_v01.cdf')
        self.assertTrue(len(files) == 1)
        self.assertTrue(files[0] == os.path.join(os.getcwd(), 'psp_data.cdf'))

    def test_local_path(self):
        # specifying local_path changes the local data directory
        files = download(local_path='psp_data/spc/l3', remote_file='https://spdf.gsfc.nasa.gov/pub/data/psp/sweap/spc/l3/l3i/2019/psp_swp_spc_l3i_20190401_v01.cdf')
        self.assertTrue(len(files) == 1)
        self.assertTrue(files[0] == os.path.join('psp_data/spc/l3', 'psp_swp_spc_l3i_20190401_v01.cdf'))


    def test_force_download(self):
        # specifying both remote_path and remote_file saves the files to the current working directory + the path specified in remote_file
        files = download(remote_path='https://spdf.gsfc.nasa.gov/pub/data/', remote_file='themis/tha/l1/state/2007/tha_l1_state_20070217_v01.cdf')
        self.assertTrue(len(files) == 1)
        self.assertTrue(files[0] == os.path.join(os.getcwd(), 'themis/tha/l1/state/2007/tha_l1_state_20070217_v01.cdf'))
        # Download the same file without force_download, should not re-download
        with self.assertLogs(level='INFO') as log:
            files = download(remote_path='https://spdf.gsfc.nasa.gov/pub/data/',
                             remote_file='themis/tha/l1/state/2007/tha_l1_state_20070217_v01.cdf')
            self.assertIn("File is current", log.output[0])
        # Download the same file with force_download, should re-download
        with self.assertLogs(level='INFO') as log:
            files = download(remote_path='https://spdf.gsfc.nasa.gov/pub/data/',
                             remote_file='themis/tha/l1/state/2007/tha_l1_state_20070217_v01.cdf',
                             force_download=True)
            self.assertIn("Downloading", log.output[0])

if __name__ == '__main__':
    unittest.main()


from pyspedas.projects.themis.load import load


def ask(site=None,
         trange=['2007-03-23', '2007-03-24'],
         group=None,
         level='l1',
         datatype='asf',
         suffix='',
         get_support_data=False,
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False):
    """
    This function loads THEMIS All Sky Keograms

    Parameters
    ----------
        site: str
            Observatory name, IF set, loads ASF data, not ASK
            Default: None, all sites are loaded
            
        trange: list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2007-03-23', '2007-03-24']

        level: str
            Data level; Valid options: 'l1'
            Default: 'l1'

        suffix: str
            The tplot variable names will be given this suffix.
            Default: no suffix is added

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  
            Default: False; only loads data with a "VAR_TYPE" attribute of "data"

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.
            Default: None; all variables are loaded

        varnames: list of str
            List of variable names to load
            Default: Empty list, so all data variables are loaded

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified
            in the trange keyword
            Default: False

    Returns
    -------
        List of str
        List of tplot variables created
        Empty list if no data
    
    Example
    -------
        import pyspedas
        from pytplot import tplot
        >>> ask_vars = pyspedas.projects.themis.ask(trange=['2013-11-05', '2013-11-06'])
        >>> tplot(['thg_ask_atha', 'thg_ask_chbg'])
    """

    return load(instrument='ask', trange=trange, level=level, datatype=datatype,
                suffix=suffix, get_support_data=get_support_data,
                varformat=varformat, varnames=varnames, stations=site,
                downloadonly=downloadonly, notplot=notplot, 
                time_clip=time_clip, no_update=no_update)




"""Automated tests for the analysis functions."""

import unittest
from unittest.mock import patch
import sys
import pyspedas
from pytplot import smooth
from pyspedas import (subtract_average, subtract_median, tsmooth, avg_data,
                      yclip, time_clip, deriv_data, tdeflag, clean_spikes,
                      tinterpol, tvectot, wavelet, time_domain_filter)
from pytplot import tcrossp
from pytplot import tdotp
from pytplot import tnormalize
from pytplot import get_data, store_data, replace_data, time_string, time_float, data_exists, del_data

import numpy as np
import copy
import logging

class BaseTestCase(unittest.TestCase):
    """Data to be used in tests."""

    def setUp(self):
        """Create a tplot variable to be used in tests."""
        store_data('test', data={'x': [1., 2., 3., 4., 5., 6.],
                                 'y': [3., 5., 8., 15., 20., 1.]})

    def tearDown(self):
        """ Delete the tplot variable."""
        del_data('*')


class AnalysisTestCases(BaseTestCase):
    """Test functions under analysis folder."""

    def test_tdotp(self):
        store_data('var1', data={'x': [0], 'y': [[3, -3, 1]]})
        store_data('var2', data={'x': [0], 'y': [[4, 9, 2]]})
        dp = tdotp('var1', 'var2')
        dpdata = get_data('var1_dot_var2')
        self.assertTrue(dpdata.y == np.array([-13]))
        dp = tdotp('var1_doesnt_exist', 'var2')
        dp = tdotp('var1', 'var2_doesnt_exist')

    def test_tcrossp(self):
        """ cross product tests"""
        cp = tcrossp([3, -3, 1], [4, 9, 2], return_data=True)
        self.assertTrue(cp.tolist() == [-15, -2, 39])
        cp = tcrossp([3, -3, 1], [4, 9, 2])
        cp = get_data(cp)
        self.assertTrue(cp.y[0, :].tolist() == [-15, -2, 39])
        store_data('var1', data={'x': [0], 'y': [[3, -3, 1]]})
        store_data('var2', data={'x': [0], 'y': [[4, 9, 2]]})
        cp = tcrossp('var1', 'var2', return_data=True)
        self.assertTrue(cp[0].tolist() == [-15, -2, 39])
        cp = tcrossp('var1', 'var2', newname='test_crossp')
        cp = get_data('test_crossp')
        self.assertTrue(cp.y[0, :].tolist() == [-15, -2, 39])

    def test_tnormalize(self):
        """ tests for normalizing tplot variables"""
        store_data('test_tnormalize', data={'x': [1, 2, 3, 4, 5], 'y': [[3, 2, 1], [1, 2, 3], [10, 5, 1], [8, 10, 14], [70, 20, 10]]})
        norm = tnormalize('test_tnormalize')
        normalized_data = get_data(norm)
        self.assertTrue(np.round(normalized_data.y[0, :], 4).tolist() == [0.8018, 0.5345, 0.2673])
        self.assertTrue(np.round(normalized_data.y[1, :], 4).tolist() == [0.2673, 0.5345, 0.8018])
        self.assertTrue(np.round(normalized_data.y[2, :], 4).tolist() == [0.8909, 0.4454, 0.0891])
        self.assertTrue(np.round(normalized_data.y[3, :], 4).tolist() == [0.4216, 0.527, 0.7379])
        self.assertTrue(np.round(normalized_data.y[4, :], 4).tolist() == [0.9526, 0.2722, 0.1361])

    def test_subtract_median(self):
        """Test subtract_median."""
        subtract_median('aaabbbccc')  # Test non-existent name
        subtract_median('test')
        d = get_data('test-m')
        self.assertTrue(d[1].tolist() == [-3.5, -1.5,  1.5,  8.5, 13.5, -5.5])
        dn = [[3., 5., 8.], [15., 20., 1.], [3., 5., 8.], [15., 20., 1.],
              [23., 15., 28.], [15., 20., float('nan')]]
        store_data('test1', data={'x': [1., 2., 3., 4., 5., 6.], 'y': dn})
        subtract_median('aaabbbcc')
        subtract_median('test1', newname='aabb')
        d = get_data('aabb')
        self.assertTrue(len(d[1]) == 6)
        subtract_median(['test', 'aabb'], newname='aaabbb')
        subtract_median('test1', overwrite=True)
        subtract_average('test', newname="testtest")
        subtract_average(['test-m', 'test'], newname="testtest2")

    def test_subtract_average(self):
        """Test subtract_average."""
        subtract_average('aaabbbccc')  # Test non-existent name
        subtract_average('test')
        d = get_data('test-d')
        self.assertTrue((np.round(d[1].tolist()) == [-6., -4., -1.,
                         6., 11., -8.]).all())
        # Test with integer data
        # int_data=[1,2,3,4,5,6]
        # store_data('test-int',data={'x':d[0],'y':int_data})
        # subtract_average('test-int')
        # d=get_data('test-int-d')
        #self.assertEqual(d.y[0],-2.5)
        dn = [[3., 5., 8.], [15., 20., 1.], [3., 5., 8.], [15., 20., 1.],
              [23., 15., 28.], [15., 20., float('nan')]]
        store_data('test1', data={'x': [1., 2., 3., 4., 5., 6.], 'y': dn})
        subtract_average('aaabbbcc')
        subtract_average('test1', newname='aabb')
        d = get_data('aabb')
        subtract_average(['test', 'aabb'], newname='aaabbb')
        subtract_average('test1', overwrite=True)
        subtract_average('test1', newname="testtest")
        subtract_average(['test1', 'test'], newname="testtest2")
        self.assertTrue(len(d[1]) == 6)

    def test_subtract_average_nan(self):
        """Test subtract_average with NaN values."""
        # Create a tplot variable with NaN values
        store_data('test-nan', data={'x': [1., 2., 3., 4.], 'y': [3., 5., np.nan, 8.]})

        # Run subtract_average
        tvar = subtract_average('test-nan')
        d = get_data(tvar[0])

        # Check that the result is not all NaN
        self.assertFalse(np.isnan(d[1]).all(), "The result should not be all NaN.")

        # Check that NaN remains where it was
        self.assertTrue(np.isnan(d[1][2]), "NaN values should remain in the output.")

        # Check that non-NaN values are properly adjusted
        expected_values = np.array([3., 5., np.nan, 8.]) - np.nanmean(np.array([3., 5., np.nan, 8.]))
        np.testing.assert_array_almost_equal(d[1], expected_values,
                                             err_msg="Non-NaN values should be adjusted by subtracting the mean.")

        # Additional NaN scenarios
        store_data('test-nan-all', data={'x': [1., 2., 3., 4.], 'y': [np.nan, np.nan, np.nan, np.nan]})

        # Check all NaN case
        result_all = subtract_average('test-nan-all')
        d_all = get_data(result_all[0])
        self.assertTrue(np.isnan(d_all[1]).all(), "All NaN input should result in all NaN output.")

    def test_subtract_median_nan(self):
        """Test subtract_median with NaN values."""
        # Create a tplot variable with NaN values
        store_data('test-nan', data={'x': [1., 2., 3., 4.], 'y': [3., 5., np.nan, 8.]})

        # Run subtract_median
        tvar = subtract_median('test-nan')
        d = get_data(tvar[0])

        # Check that the result is not all NaN
        self.assertFalse(np.isnan(d[1]).all(), "The result should not be all NaN.")

        # Check that NaN remains where it was
        self.assertTrue(np.isnan(d[1][2]), "NaN values should remain in the output.")

        # Check that non-NaN values are properly adjusted by subtracting the median
        expected_values = np.array([3., 5., np.nan, 8.]) - np.nanmedian(np.array([3., 5., np.nan, 8.]))
        np.testing.assert_array_almost_equal(d[1], expected_values,
                                             err_msg="Non-NaN values should be adjusted by subtracting the median.")

        # Additional NaN scenarios
        store_data('test-nan-all', data={'x': [1., 2., 3., 4.], 'y': [np.nan, np.nan, np.nan, np.nan]})

        # Check all NaN case
        result_all = subtract_median('test-nan-all')
        d_all = get_data(result_all[0])
        self.assertTrue(np.isnan(d_all[1]).all(), "All NaN input should result in all NaN output.")

    def test01_subtract_average_with_suffix(self):
        """Test subtract_average with suffix."""
        subtract_average('test', suffix='-sfx')
        self.assertIsNotNone(get_data('test-sfx'))

    def test02_subtract_average_with_newname_and_suffix(self):
        """Test subtract_average with both custom suffix and newname."""
        subtract_average('test', newname='new_test', suffix='-sfx')
        self.assertIsNotNone(get_data('new_test'))
        # Ensure suffix is not applied when newname is provided
        self.assertIsNone(get_data('test-sfx'))

    def test_subtract_average_with_invalid_names(self):
        """Test subtract_average with invalid names input."""
        result = subtract_average('invalid_name')
        self.assertIsNone(result)  # Should handle gracefully

    def test_subtract_average_with_empty_names(self):
        """Test subtract_average with empty names list."""
        result = subtract_average([])
        self.assertIsNone(result)  # Should handle gracefully

    def test_subtract_average_all_same_values(self):
        """Test subtract_average with all same values in the dataset."""
        store_data('test_same', data={'x': [1., 2., 3.], 'y': [5., 5., 5.]})
        subtract_average('test_same')
        d = get_data('test_same-d')
        self.assertTrue((d[1] == [0, 0, 0]).all())

    def test_subtract_average_one_element(self):
        """Test subtract_average with dataset containing only one element."""
        store_data('test_one', data={'x': [1.], 'y': [10.]})
        subtract_average('test_one')
        d = get_data('test_one-d')
        self.assertTrue((d[1] == [0]).all())

    def test_subtract_average_logging_deprecation(self):
        """Test logging for deprecation warning with new_names."""
        with self.assertLogs(level='INFO') as log:
            subtract_average('test', new_names='test_deprecated')
            self.assertIn("The new_names parameter is deprecated", log.output[0])

    # For subtract_median

    def test_subtract_median_basic(self):
        """Test basic functionality of subtract_median."""
        store_data('test', data={'x': [1., 2., 3.], 'y': [4., 5., 6.]})
        subtract_median('test')
        d = get_data('test-m')
        self.assertIsNotNone(d)

    @unittest.skip('The architecture of pytplot does not support the following mock')
    def test_subtract_median_parameter_passing(self):
        """Test that parameters are correctly passed to subtract_average via subtract_median."""

        #with patch('pytplot.tplot_math.subtract_average') as mock_subtract_average:
        with patch('pytplot.tplot_math.subtract_average.subtract_average') as mock_subtract_average:
        #with patch('pyspedas.subtract_average') as mock_subtract_average:
            print(mock_subtract_average)
            subtract_median('test', newname='new_test', suffix='-sfx', overwrite=True)

            # Check that subtract_average was called with the correct parameters, including median=1
            mock_subtract_average.assert_called_once_with(
                'test', newname='new_test', suffix='-sfx', overwrite=True, median=1
            )


    def test_subtract_median_deprecation_warning(self):
        """Test subtract_median for deprecation warning with new_names."""
        with self.assertLogs(level='INFO') as log:
            subtract_median('test', new_names='deprecated_test')
            self.assertIn("The new_names parameter is deprecated", log.output[0])


    def test_yclip(self):
        """Test yclip."""
        yclip('aaabbbccc', 0.0, 12.0)  # Test non-existent name
        yclip('aabb', 0.0, 12.0)
        yclip('test', 0.0, 12.0)
        d = get_data('test-clip')
        # Replace nan with -99.0
        dd = np.nan_to_num(d[1], nan=-99.)
        yclip('test', 0.0, 12.0, newname='name-clip')
        yclip(['test', 'name-clip'], 0.0, 12.0, newname='name1-clip')
        yclip('test', 0.0, 12.0, overwrite=True)
        yclip('test', 0.0, 12.0, newname="testtest")
        yclip(['test', 'test-clip'], 0.0, 12.0, newname="testtest2")
        self.assertTrue((dd == [3., 5., 8., -99., -99., 1.]).all())

    def test_timeclip(self):
        """Test time_clip."""
        time_clip('aaabbbccc', 1577308800, 1577598800)  # Test non-existent
        tn = [1577112800, 1577308800, 1577598800, 1577608800, 1577998800,
              1587998800]
        dn = [3., 5., 8., 15., 20., 1.]
        store_data('test1', data={'x': tn, 'y': dn})
        time_clip('aaabbb', 1577308800, 1577598800)
        time_clip('test1', 1577112800, 1577608800)
        d = get_data('test1-tclip')
        dd = d[1]
        time_clip('test', 1577308800, 1577598800, newname='name-clip')
        time_clip(['test', 'name-clip'], 1577308800, 1577598800,
                  newname='name1-ci')
        time_clip('test', 1577308800, 1577598800, overwrite=True)
        time_clip('test', 1577308800, 1577598800, newname="testtest")
        time_clip(['test', 'test1'], 1577308800, 1577598800,
                  newname="testtest2")
        time_clip('test1', 1677112800, 1577608800)
        self.assertTrue((dd == [3., 5., 8., 15.]).all())

    def test_avg_data(self):
        """Test avg_data."""
        avg_data('aaabbbccc', width=2)  # Test non-existent name
        avg_data('test', width=2)
        d = get_data('test-avg')
        self.assertTrue((d[1] == [4.0, 11.5, 10.5]).all())
        avg_data('test', width=2, overwrite=True)  # Test overwrite
        store_data('test', data={'x': [1., 2., 3., 4., 5., 6.],
                                 'y': [3., 5., 8., -4., 20., 1.]})
        avg_data('test', width=2, newname='aabb')  # Test new_names
        d = get_data('aabb')
        # Test multiple names
        avg_data(['test', 'aabb'], newname='aaabbb', width=2)
        dt = [1., 12., 13., 14., 15., 16.]
        dn = [[3., 5., 8.], [15., 20., 1.], [3., 5., 8.], [15., 20., 1.],
              [23., 15., 28.], [15., 20., 1.]]
        dv = dn
        store_data('test1', data={'x': dt, 'y': dn, 'v': dv})
        avg_data('test1', width=2)  # Test 3-d data
        avg_data('test1', newname='test2', res=2.)  # Test a reasonable resolution
        avg_data('test1', res=-1.)  # Test res error
        avg_data('test1', res=1.e8)  # Test res error
        d2 = get_data('test2')
        self.assertTrue(len(d2) > 0)
        self.assertTrue(d2[1][-1][0] == 19.0)
        self.assertTrue(len(d2[2]) == len(d2[0]))

    def test_clean_spikes(self):
        """Test clean_spikes."""
        clean_spikes('aaabbbccc', nsmooth=3)  # Test non-existent name
        clean_spikes('test', nsmooth=3)
        d = get_data('test-despike')
        self.assertTrue(len(d[1]) == 6)
        # Now test 3 dim data.
        dn = [[3., 5., 8.], [15., 20., 1.], [3., 5., 8.], [15., 20., 1.],
              [23., 15., 28.], [15., 20., 1.]]
        store_data('test1', data={'x': [1., 2., 3., 4., 5., 6.], 'y': dn})
        clean_spikes('test1', nsmooth=3)
        d2 = get_data('test1-despike')
        clean_spikes('test', newname='test_desp', nsmooth=3, sub_avg=True)
        clean_spikes(['test', 'test1'], newname='test1-desp')
        clean_spikes('test1', overwrite=True)
        self.assertTrue(len(d2[1]) == 6)

    def test_tdeflag(self):
        """Test tdeflag."""
        tdeflag('aaabbbccc')  # Test non-existent name
        dn = [3., float('NaN'), 8., float('NaN'), 20., 1.]
        len_dn = len(dn)
        replace_data('test', dn)
        tdeflag('test')
        d = get_data('test-deflag')
        tdeflag('test', overwrite=True)
        tdeflag('test', newname="testtest")
        tdeflag(['test', 'test-deflag'], newname="testtest2")
        # Length should be two less, because NaNs were removed.
        self.assertTrue(len(d[1]) == len_dn - 2)

    def test_tdeflag_2d(self):
        """Test tdeflag."""
        times=np.array([1.0,2.0,3.0,4.0,5.0])
        vals = np.array([ [1.0, 2.0, 3.0],
                          [1.1, 2.1, 3.1],
                          [1.2, 2.2, 3.2],
                          [1.3, 2.3, 3.3],
                          [1.4, 2.4, 3.4]])
        dn = [3., float('NaN'), 8., float('NaN'), 20., 1.]
        
        vals_noflag = copy.deepcopy(vals)
        store_data('vals_noflag',data={'x':times, 'y':vals_noflag})
        tdeflag('vals_noflag',newname='vals_noflag_remove') # no method should default to remove_nan
        tdeflag('vals_noflag',newname='vals_noflag_linear', method='linear')
        tdeflag('vals_noflag',newname='vals_noflag_repeat', method='repeat')
        tdeflag('vals_noflag',newname='vals_noflag_replace', method='replace', fillval=100.0)
        d1=get_data('vals_noflag_remove')
        d2=get_data('vals_noflag_linear')
        d3=get_data('vals_noflag_repeat')
        d4=get_data('vals_noflag_replace')
        d5=get_data('vals_noflag_linear')
        np.testing.assert_array_equal(d1.y,vals_noflag)
        np.testing.assert_array_equal(d2.y,vals_noflag)
        np.testing.assert_array_equal(d3.y,vals_noflag)
        np.testing.assert_array_equal(d4.y,vals_noflag)
        np.testing.assert_array_equal(d5.y,vals_noflag)

        # There can be multiple flags. Make sure the 'no unflagged data' case is
        # properly defected when each individual flag leaves some valid data, but
        # the combination flags everything.

        # commented out for now until they pass
        # tdeflag('vals_noflag', newname='vals_all_multiflag_linear', method='linear', flag=[2.0,2.1,2.3,2.4,2.5])
        # self.assertFalse(data_exists('vals_all_multiflag_linear'))
        # tdeflag('vals_noflag', newname='vals_all_multiflag_repeat', method='repeat', flag=[2.0,2.1,2.3,2.4,2.5])
        # self.assertFalse(data_exists('vals_all_multiflag_repeat'))

        vals_flag_beginning = copy.deepcopy(vals)
        vals_flag_beginning[0,1] = np.nan
        store_data('vals_flag_beginning',data={'x':times, 'y':vals_flag_beginning})        
        tdeflag('vals_flag_beginning',newname='vals_flag_beginning_remove', method='remove_nan')
        tdeflag('vals_flag_beginning',newname='vals_flag_beginning_linear', method='linear')
        tdeflag('vals_flag_beginning',newname='vals_flag_beginning_repeat', method='repeat')
        tdeflag('vals_flag_beginning',newname='vals_flag_beginning_replace', method='replace', fillval=100.0)
        d1=get_data('vals_flag_beginning_remove')
        d2=get_data('vals_flag_beginning_linear')
        d3=get_data('vals_flag_beginning_repeat')
        d4=get_data('vals_flag_beginning_replace')
        np.testing.assert_equal(len(d1.times),4)
        np.testing.assert_equal(d1.times[0],times[1])
        np.testing.assert_equal(d2.y[0, 1],2.1)
        np.testing.assert_equal(d3.y[0, 1], 2.1)
        np.testing.assert_equal(d4.y[0, 1], 100.0)

        vals_flag_end = copy.deepcopy(vals)
        vals_flag_end[4,1] = np.nan
        store_data('vals_flag_end',data={'x':times, 'y':vals_flag_end})        
        tdeflag('vals_flag_end',newname='vals_flag_end_remove', method='remove_nan')
        tdeflag('vals_flag_end',newname='vals_flag_end_linear', method='linear')
        tdeflag('vals_flag_end',newname='vals_flag_end_repeat', method='repeat')
        tdeflag('vals_flag_end',newname='vals_flag_end_replace', method='replace', fillval=100.0)
        d1=get_data('vals_flag_end_remove')
        d2=get_data('vals_flag_end_linear')
        d3=get_data('vals_flag_end_repeat')
        d4=get_data('vals_flag_end_replace')
        np.testing.assert_equal(len(d1.times),4)
        np.testing.assert_equal(d1.times[3],times[3])
        np.testing.assert_equal(d2.y[4, 1],2.3)
        np.testing.assert_equal(d3.y[4, 1], 2.3)
        np.testing.assert_equal(d4.y[4, 1], 100.0)

        vals_flag_middle = copy.deepcopy(vals)
        vals_flag_middle[2, 1] = np.nan
        store_data('vals_flag_middle',data={'x':times, 'y':vals_flag_middle})        
        tdeflag('vals_flag_middle',newname='vals_flag_middle_remove', method='remove_nan')
        tdeflag('vals_flag_middle',newname='vals_flag_middle_linear', method='linear')
        tdeflag('vals_flag_middle',newname='vals_flag_middle_repeat', method='repeat')
        tdeflag('vals_flag_middle',newname='vals_flag_middle_replace', method='replace', fillval=100.0)
        d1=get_data('vals_flag_middle_remove')
        d2=get_data('vals_flag_middle_linear')
        d3=get_data('vals_flag_middle_repeat')
        d4=get_data('vals_flag_middle_replace')
        np.testing.assert_equal(len(d1.times),4)
        np.testing.assert_equal(d1.times[2],times[3])
        np.testing.assert_equal(d2.y[2, 1],2.2)
        np.testing.assert_equal(d3.y[2, 1], 2.1)
        np.testing.assert_equal(d4.y[2, 1], 100.0)

        vals_flag_all = copy.deepcopy(vals)
        vals_flag_all[:,:] = np.nan
        store_data('vals_flag_all',data={'x':times, 'y':vals_flag_all})
        tdeflag('vals_flag_all',newname='vals_flag_all_remove', method='remove_nan')
        self.assertFalse(data_exists('vals_flag_all_remove'))
        tdeflag('vals_flag_all',newname='vals_flag_all_linear', method='linear')
        self.assertFalse(data_exists('vals_flag_all_linear'))
        tdeflag('vals_flag_all',newname='vals_flag_all_repeat', method='repeat')
        self.assertFalse(data_exists('vals_flag_all_repeat'))
        tdeflag('vals_flag_all',newname='vals_flag_all_replace', method='replace', fillval=100.0)
        d=get_data('vals_flag_all_replace')
        self.assertEqual(d.y[0, 0], 100.0)

    def test_tdeflag_1d(self):
        times = [1.0, 2.0, 3.0, 4.0, 5.0]
        data = [1.0, 2.0, 3.0, 4.0, 5.0]
        vals_noflag = copy.deepcopy(data)
        store_data('vals_noflag_1d', data={'x':times, 'y':vals_noflag})
        tdeflag('vals_noflag_1d', newname='vals_noflag_1d_linear', method='linear')
        d1=get_data('vals_noflag_1d_linear')
        np.testing.assert_array_equal(d1.y, vals_noflag)

    def test_deriv_data(self):
        """Test deriv_data."""
        deriv_data('aaabbbccc')  # Test non-existent name
        deriv_data('test')
        d = get_data('test-der')
        deriv_data('test', overwrite=True)
        deriv_data('test', newname="testtest")
        deriv_data(['test', 'test-der'], newname="testtest2")
        self.assertTrue((d[1] == [2., 2.5, 5.,   6., -7., -19.]).all())

    def test_tvectot(self):
        from pyspedas.projects.themis import state
        from pytplot import data_exists
        state(probe='a')
        tvectot('tha_pos', join_component=True)
        self.assertTrue(data_exists('tha_pos_tot'))
        d = get_data('tha_pos_tot')
        s = d.y.shape[1]
        self.assertEqual(s,4)
        tvectot('tha_pos',newname='tha_pos_total')
        self.assertTrue(data_exists('tha_pos_total'))
        tvectot('tha_pos',suffix='_rtot')
        self.assertTrue(data_exists('tha_pos_rtot'))

    def test_wavelet(self):
        # Create a tplot variable that contains a wave.
        t = np.arange(4000.)
        y = np.sin(2*np.pi*t/32.)
        y2 = np.sin(2*np.pi*t/64.)
        y[1000:3000] = y2[1000:3000]
        var = 'sin_wav'
        time = time_float('2010-01-01') + 10*t
        store_data(var, data={'x':time, 'y':y})

        # Gaussian Derivative wavelets transformation.
        powervar = wavelet(var, wavename='gaus1')
        pvar = powervar[0]
        self.assertTrue(data_exists(pvar))

    def test_time_domain_filter(self):
        # Create a tplot variable that contains a wave.
        t = np.arange(4000.)
        y = np.sin(2*np.pi*t/32.)
        y2 = np.sin(2*np.pi*t/64.)
        y[1000:3000] = y2[1000:3000]
        dat = np.zeros((4000,3))
        dat[:,0] = y
        dat[:,1] = y
        dat[:,2] = y
        var = 'sin_wav'
        time = time_float('2010-01-01') + 10*t
        output = time_domain_filter(dat,time,16.0, 48.0)
        self.assertEqual(output.shape[0], dat.shape[0])
        self.assertEqual(output.shape[1], dat.shape[1])


    def test_tsmooth(self):
        """Test smooth."""
        tsmooth('aaabbbccc')  # Test non-existent name
        a = [1.0, 1.0, 2.0, 3.0, 4.0, 1.0, 4.0, 3.0, 2.0, 1.0, 1.0]
        x = smooth(a, 3)
        r = [1.0, 1.3333333333333333, 2.0, 3.0, 2.6666666666666665,
             3.0, 2.6666666666666665, 3.0, 2.0, 1.3333333333333333, 1.0]
        self.assertTrue(x == r)
        b = [1.0, 1.0, 2.0, 3.0, np.nan, np.nan, np.nan, np.nan, 2.0, 1.0, 1.0]
        y = smooth(b, width=3)
        ry = [1.0, 1.3333333333333333, 2.0, 1.6666666666666665, 1.0,
              np.nan, np.nan, 0.6666666666666666, 1.0, 1.3333333333333333, 1.0]
        self.assertTrue(y == ry)
        tsmooth('test')
        d = get_data('test-s')
        tsmooth('test', overwrite=True)
        tsmooth('test', newname="testtest")
        tsmooth(['test', 'test-s'], newname="testtest2")
        self.assertTrue(d[1].tolist() == [3.,  5.,  8., 15., 20.,  1.])

    def test_tinterpol(self):
        """Test tinterpol."""
        tinterpol('aaabbbccc', 'test')  # Test non-existent name
        tn = [1., 1.5, 4.6, 5.8, 6.]
        dn = [10., 15., 46., 58., 60.]
        data = np.array([[0, 1, 2, 3, 4],
                        [5, 6, 7, 8, 9],
                        [10, 11, 12, 13, 14],
                        [15, 16, 17, 18, 19],
                        [20, 21, 22, 23, 24]])
        store_data('test1', data={'x': tn, 'y': dn})
        store_data('test2', data={'x': tn, 'y': data, 'v': [10, 20, 30, 40, 50]})
        tinterpol('test1', 'test')
        tinterpol('test1', 'doesnt_exist')
        tinterpol('test2', 'test', newname='')
        tinterpol('test2', [1, 2, 3, 4, 5, 6], newname='pyarray_float')
        tinterpol('test2', np.array([1, 2, 3, 4, 5, 6]), newname='nparray_float')
        tinterpol('test2', time_string([1,2,3,4,5,6]), newname='pyarray_str')
        tinterpol('test2', np.array(time_string([1,2,3,4,5,6])), newname='nparray_str')
        bad_datatype = {}
        tinterpol('test2', [bad_datatype], newname='nparray_bad_datatype')

        d = get_data('test1-itrp')
        self.assertTrue(d[1][1] == 20.)
        d2 = get_data('pyarray_float')
        self.assertTrue(abs(d2[1][1][0] - 5.80645161) < 1e-6)
        d3 = get_data('nparray_float')
        self.assertTrue(abs(d3[1][1][0] - 5.80645161) < 1e-6)
        d4 = get_data('pyarray_str')
        self.assertTrue(abs(d3[1][1][0] - 5.80645161) < 1e-6)
        d5 = get_data('nparray_str')
        self.assertTrue(abs(d3[1][1][0] - 5.80645161) < 1e-6)

    def test_scipy_interp1d(self):
        import scipy
        import numpy as np
        time_strings_input = np.array(['2018-07-01T13:02:16.892474880',
                                       '2018-07-01T13:02:16.922475008',
                                       '2018-07-01T13:02:16.952474880'])
        values_input = np.array([0.028584518, 0., 0.013626526],dtype=np.float32)

        input_times_npdt64 = np.array([np.datetime64(t) for t in time_strings_input])
        interp_to_times_npdt64 = np.array(input_times_npdt64[1])

        input_times_float64 = input_times_npdt64.astype(np.float64)
        interp_to_time_float64 = interp_to_times_npdt64.astype(np.float64)

        interpolator = scipy.interpolate.interp1d(input_times_float64, values_input, kind='linear')
        result=interpolator(interp_to_time_float64)
        print(result)
        # Known to fail.  This affects xarray.interp and the current version of tinterpol.
        #self.assertTrue((result >= 0.0).all())

    def test_scipy_spline(self):
        import scipy
        import numpy as np
        time_strings_input = np.array(['2018-07-01T13:02:16.892474880',
                                       '2018-07-01T13:02:16.922475008',
                                       '2018-07-01T13:02:16.952474880'])
        values_input = np.array([0.028584518, 0., 0.013626526],dtype=np.float32)

        input_times_npdt64 = np.array([np.datetime64(t) for t in time_strings_input])
        interp_to_times_npdt64 = np.array(input_times_npdt64[1])

        input_times_float64 = input_times_npdt64.astype(np.float64)
        interp_to_time_float64 = interp_to_times_npdt64.astype(np.float64)

        interpolator = scipy.interpolate.make_interp_spline(input_times_float64, values_input, k=1)
        result=interpolator(interp_to_time_float64)
        print(result)
        # make_interp_spline() with k=1 gives the expected result
        self.assertTrue((result >= 0.0).all())

    def test_xarray_interp(self):
        import xarray as xr
        import numpy as np

        time_strings_input = np.array(['2018-07-01T13:02:16.892474880',
                                       '2018-07-01T13:02:16.922475008',
                                       '2018-07-01T13:02:16.952474880'])
        values_input = np.array([0.028584518, 0., 0.013626526],dtype=np.float32)

        input_times_npdt64 = np.array([np.datetime64(t) for t in time_strings_input])
        interp_to_times_npdt64 = np.array(input_times_npdt64[1])

        data_array = xr.DataArray(values_input,dims=['time'],coords={'time':('time',input_times_npdt64)})

        result = data_array.interp({"time": interp_to_times_npdt64},method='linear')
        # This is known to fail, due to issues in scipy.interpolate.interp1d
        print(result.values)
        # result.values is [-3.469446951953614e-18]
        #self.assertTrue((result.values >= 0.0).all())

    def test_xarray_interp_float_times(self):
        import xarray as xr
        import numpy as np

        time_strings_input = np.array(['2018-07-01T13:02:16.892474880',
                                       '2018-07-01T13:02:16.922475008',
                                       '2018-07-01T13:02:16.952474880'])
        values_input = np.array([0.028584518, 0., 0.013626526],dtype=np.float32)

        input_times_npdt64 = np.array([np.datetime64(t) for t in time_strings_input])
        interp_to_times_npdt64 = np.array(input_times_npdt64[1])

        input_times_float64 = input_times_npdt64.astype(np.float64)
        interp_to_time_float64 = interp_to_times_npdt64.astype(np.float64)

        data_array = xr.DataArray(values_input,dims=['time'],coords={'time':('time',input_times_float64)})

        result = data_array.interp({"time": interp_to_time_float64},method='linear')
        # This is known to fail, due to issues in scipy.interpolate.interp1d
        print(result.values)
        # result.values is [-3.469446951953614e-18]
        #self.assertTrue((result.values >= 0.0).all())

    def test_numpy_interp(self):
        time_strings_input = np.array(['2018-07-01T13:02:16.892474880',
                                       '2018-07-01T13:02:16.922475008',
                                       '2018-07-01T13:02:16.952474880'])
        values_input = np.array([0.028584518, 0., 0.013626526],dtype=np.float32)

        time_strings_interp_to = np.array(['2018-07-01T13:02:16.922475008'])
        input_times_npdt64 = np.array([np.datetime64(t) for t in time_strings_input])
        interp_to_times_npdt64 = np.array([np.datetime64(t) for t in time_strings_interp_to])
        input_times_float64 = input_times_npdt64.astype(np.float64)
        interp_to_time_float64 = interp_to_times_npdt64.astype(np.float64)
        result=np.interp(interp_to_time_float64,input_times_float64,values_input)
        # This works, unlike scipy and xarray!
        self.assertTrue((result >= 0.0).all())


    def test_tinterpol_nonnegative2(self):
        time_strings_input = np.array(['2018-07-01T13:02:16.892474880',
                                       '2018-07-01T13:02:16.922475008',
                                       '2018-07-01T13:02:16.952474880'])
        values_input = np.array([0.028584518, 0., 0.013626526],dtype=np.float32)
        time_strings_interp_to = np.array(['2018-07-01T13:02:16.922475008'])
        input_times_npdt64 = np.array([np.datetime64(t) for t in time_strings_input])
        interp_to_times_npdt64 = np.array([np.datetime64(t) for t in time_strings_interp_to])
        store_data('interp_input', data={'x':input_times_npdt64, 'y':values_input})
        store_data('interp_to', data={'x':interp_to_times_npdt64, 'y':[0.0]})
        tinterpol('interp_input', 'interp_to', newname='interp_result')
        result=get_data('interp_result')
        # This is known to fail, apparently due to limitations of scipy.interpolate.interp1d which are
        # unlikely to ever be fixed. See tests above for xarray.interp and scipy
        # self.assertTrue((result.y >= 0.0).all())

    def test_tinterpol_slinear(self):
        # This test uses the 'slinear' interpolation method (order 1 spline) which
        # seems not to be susceptible to the issue that method='linear' has with interpolating
        # to points exactly matching the input times.
        time_strings_input = np.array(['2018-07-01T13:02:16.892474880',
                                       '2018-07-01T13:02:16.922475008',
                                       '2018-07-01T13:02:16.952474880'])
        values_input = np.array([0.028584518, 0., 0.013626526],dtype=np.float32)
        time_strings_interp_to = np.array(['2018-07-01T13:02:16.922475008'])
        input_times_npdt64 = np.array([np.datetime64(t) for t in time_strings_input])
        interp_to_times_npdt64 = np.array([np.datetime64(t) for t in time_strings_interp_to])
        store_data('interp_input', data={'x':input_times_npdt64, 'y':values_input})
        store_data('interp_to', data={'x':interp_to_times_npdt64, 'y':[0.0]})
        tinterpol('interp_input', 'interp_to', newname='interp_result', method='slinear')
        result=get_data('interp_result')
        self.assertTrue((result.y >= 0.0).all())

if __name__ == '__main__':
    unittest.main()
